# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations
import mptt.fields
from django.conf import settings
from django.db.models import CASCADE


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='Handler',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('date_created', models.DateTimeField(auto_now_add=True, verbose_name='Date Created')),
                ('date_updated', models.DateTimeField(auto_now=True, verbose_name='Date Updated')),
                ('hash', models.CharField(unique=True, max_length=200, verbose_name='Hash')),
                ('method', models.CharField(max_length=200, verbose_name='Callback Method')),
                ('handler_cls', models.CharField(max_length=200, verbose_name='HandlerClass')),
                ('enabled', models.BooleanField(default=True, verbose_name='Enabled')),
            ],
            options={
                'verbose_name': 'Handler',
                'verbose_name_plural': 'Handlers',
            },
        ),
        migrations.CreateModel(
            name='Proceeding',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('date_created', models.DateTimeField(auto_now_add=True, verbose_name='Date Created')),
                ('date_updated', models.DateTimeField(auto_now=True, verbose_name='Date Updated')),
                ('object_id', models.PositiveIntegerField(verbose_name='Related Object')),
                ('field', models.CharField(max_length=200, verbose_name='Field')),
                ('transaction_date', models.DateTimeField(null=True, blank=True)),
                ('status', models.IntegerField(default=0, verbose_name='Status', choices=[(0, 'Pending'), (1, 'Approved'), (2, 'Rejected')])),
                ('skip', models.BooleanField(default=False, verbose_name='Skip')),
                ('order', models.IntegerField(default=0, verbose_name='Order')),
                ('enabled', models.BooleanField(default=True, verbose_name='Enabled?')),
                ('content_type', models.ForeignKey(verbose_name='Content Type', to='contenttypes.ContentType', on_delete=CASCADE)),
                ('groups', models.ManyToManyField(to='auth.Group', verbose_name='Groups')),
            ],
            options={
                'verbose_name': 'Proceeding',
                'verbose_name_plural': 'Proceedings',
            },
        ),
        migrations.CreateModel(
            name='ProceedingMeta',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('date_created', models.DateTimeField(auto_now_add=True, verbose_name='Date Created')),
                ('date_updated', models.DateTimeField(auto_now=True, verbose_name='Date Updated')),
                ('order', models.IntegerField(default=0, verbose_name='Order')),
                ('action_text', models.TextField(max_length=200, null=True, verbose_name='Action Text', blank=True)),
                ('groups', models.ManyToManyField(to='auth.Group', verbose_name='Groups')),
                ('permissions', models.ManyToManyField(to='auth.Permission', verbose_name='Permissions')),
            ],
            options={
                'verbose_name': 'Proceeding Meta',
                'verbose_name_plural': 'Proceeding Metas',
            },
        ),
        migrations.CreateModel(
            name='ProceedingTrack',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('date_created', models.DateTimeField(auto_now_add=True, verbose_name='Date Created')),
                ('date_updated', models.DateTimeField(auto_now=True, verbose_name='Date Updated')),
                ('process_date', models.DateTimeField(auto_now_add=True, null=True)),
                ('previous_track', mptt.fields.TreeOneToOneField(related_name='next_track', null=True, blank=True, to='river.ProceedingTrack', verbose_name='Previous track', on_delete=CASCADE)),
                ('proceeding', models.ForeignKey(related_name='tracks', verbose_name='Proceeding', to='river.Proceeding', on_delete=CASCADE)),
            ],
            options={
                'verbose_name': 'Proceeding Track',
                'verbose_name_plural': 'Proceeding Tracks',
            },
        ),
        migrations.CreateModel(
            name='State',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('date_created', models.DateTimeField(auto_now_add=True, verbose_name='Date Created')),
                ('date_updated', models.DateTimeField(auto_now=True, verbose_name='Date Updated')),
                ('label', models.CharField(max_length=50)),
                ('description', models.CharField(max_length=200, null=True, verbose_name='Description', blank=True)),
            ],
            options={
                'verbose_name': 'State',
                'verbose_name_plural': 'States',
            },
        ),
        migrations.CreateModel(
            name='Transition',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('date_created', models.DateTimeField(auto_now_add=True, verbose_name='Date Created')),
                ('date_updated', models.DateTimeField(auto_now=True, verbose_name='Date Updated')),
                ('field', models.CharField(max_length=200, verbose_name='Field')),
                ('direction', models.SmallIntegerField(default=1, verbose_name='Transition Direction')),
                ('content_type', models.ForeignKey(verbose_name='Content Type', to='contenttypes.ContentType', on_delete=CASCADE)),
                ('destination_state', models.ForeignKey(related_name='transitions_as_destination', verbose_name='Next State', to='river.State', on_delete=CASCADE)),
                ('source_state', models.ForeignKey(related_name='transitions_as_source', verbose_name='Source State', to='river.State', on_delete=CASCADE)),
            ],
            options={
                'verbose_name': 'Transition',
                'verbose_name_plural': 'Transitions',
            },
        ),
        migrations.AddField(
            model_name='proceedingmeta',
            name='transition',
            field=models.ForeignKey(verbose_name='Transition', to='river.Transition', on_delete=CASCADE),
        ),
        migrations.AddField(
            model_name='proceeding',
            name='meta',
            field=models.ForeignKey(related_name='proceedings', verbose_name='Meta', to='river.ProceedingMeta', on_delete=CASCADE),
        ),
        migrations.AddField(
            model_name='proceeding',
            name='permissions',
            field=models.ManyToManyField(to='auth.Permission', verbose_name='Permissions'),
        ),
        migrations.AddField(
            model_name='proceeding',
            name='transactioner',
            field=models.ForeignKey(verbose_name='Transactioner', blank=True, to=settings.AUTH_USER_MODEL, null=True, on_delete=CASCADE),
        ),
        migrations.AlterUniqueTogether(
            name='proceedingmeta',
            unique_together=set([('transition', 'order')]),
        ),
    ]



from django.contrib.contenttypes.models import ContentType
from django.db.models import CASCADE
from django.db.models.signals import post_save

from river.utils.error_code import ErrorCode
from river.utils.exceptions import RiverException

try:
    from django.contrib.contenttypes.fields import GenericRelation
except ImportError:
    from django.contrib.contenttypes.generic import GenericRelation

from river.models.state import State
from river.models.proceeding import Proceeding
from river.services.object import ObjectService
from river.services.transition import TransitionService

__author__ = 'ahmetdal'

from django.db import models

class_field_rl = {}


def _get_cls_identifier(cls):
    return cls.__module__ + "." + cls.__name__


def _get_identifier(cls_identifier, field):
    return "%s_%s" % (cls_identifier, field)


class StateField(models.ForeignKey):
    def __init__(self, state_model=State, *args, **kwargs):
        kwargs['null'] = True
        kwargs['blank'] = True
        kwargs['to'] = '%s.%s' % (state_model._meta.app_label, state_model._meta.object_name)
        kwargs['on_delete'] = kwargs.get('on_delete', CASCADE)
        super(StateField, self).__init__(*args, **kwargs)

    def contribute_to_class(self, cls, name):
        def is_workflow_completed(workflow_object):
            return ObjectService.is_workflow_completed(workflow_object)

        def proceed(self, user, *args, **kwargs):
            TransitionService.proceed(self, user, *args, **kwargs)

        @property
        def on_initial_state(self):
            from river.services.state import StateService

            return StateService.get_initial_state(ContentType.objects.get_for_model(self)) == self.get_state()

        @property
        def on_final_state(self):
            from river.services.state import StateService

            return self.get_state() in StateService.get_final_states(ContentType.objects.get_for_model(self))

        def get_initial_state(self):
            from river.services.state import StateService

            return StateService.get_initial_state(ContentType.objects.get_for_model(self))

        def get_available_proceedings(self, *args, **kwargs):
            from river.services.proceeding import ProceedingService

            return ProceedingService.get_available_proceedings(self, [self.get_state()], *args, **kwargs)

        @property
        def initial_proceedings(self):
            from river.services.proceeding import ProceedingService

            return self.get_state() in ProceedingService.get_initial_proceedings(ContentType.objects.get_for_model(self))

        @property
        def final_proceedings(self):
            from river.services.proceeding import ProceedingService

            return self.get_state() in ProceedingService.get_final_proceedings(ContentType.objects.get_for_model(self))

        @property
        def next_proceedings(self):
            from river.services.proceeding import ProceedingService

            return self.get_state() in ProceedingService.get_next_proceedings(ContentType.objects.get_for_model(self))

        @property
        def proceeding(self):
            try:
                return self.proceedings.filter(transaction_date__isnull=False).latest('transaction_date')
            except Proceeding.DoesNotExist:
                return None

        def _get_state(self):
            return getattr(self, name)

        def _set_state(self, state):
            setattr(self, name, state)

        field_identifiers = [_get_identifier(c, f) for c, f in class_field_rl.items()]
        cls_identifier = _get_cls_identifier(cls)
        field_identifier = _get_identifier(cls_identifier, name)
        if field_identifier not in field_identifiers:

            if cls_identifier in class_field_rl.keys():
                raise RiverException(ErrorCode.MULTIPLE_STATE_FIELDS,
                                     "There can be only one state field in a model class. Class:%s - Field:%s." % (cls.__name__, name))

            class_field_rl[cls_identifier] = name

        self.model = cls
        self.__add_to_class(cls, "proceedings", GenericRelation('%s.%s' % (Proceeding._meta.app_label, Proceeding._meta.object_name)))
        self.__add_to_class(cls, "proceeding", proceeding)

        self.__add_to_class(cls, "is_workflow_completed", is_workflow_completed)
        self.__add_to_class(cls, "proceed", proceed)

        self.__add_to_class(cls, "on_initial_state", on_initial_state)
        self.__add_to_class(cls, "on_final_state", on_final_state)

        self.__add_to_class(cls, "get_initial_state", get_initial_state)
        self.__add_to_class(cls, "get_available_proceedings", get_available_proceedings)

        self.__add_to_class(cls, "initial_proceedings", initial_proceedings)
        self.__add_to_class(cls, "final_proceedings", final_proceedings)
        self.__add_to_class(cls, "next_proceedings", next_proceedings)

        self.__add_to_class(cls, "get_state", _get_state)
        self.__add_to_class(cls, "set_state", _set_state)

        super(StateField, self).contribute_to_class(cls, name)

        post_save.connect(_post_save, self.model, False, dispatch_uid='%s_%s_riverstatefield_post' % (self.model, name))

    def __add_to_class(self, cls, key, value):
        if not hasattr(cls, key):
            cls.add_to_class(key, value)


def _post_save(sender, instance, created, *args, **kwargs):  # signal, sender, instance):
    """
    Desc:  Generate TransitionProceedings according to ProceedingMeta of the content type at the beginning.
    :param kwargs:
    :return:
    """
    from river.services.state import StateService

    if created:
        ObjectService.register_object(instance)
    if not instance.get_state():
        init_state = StateService.get_initial_state(ContentType.objects.get_for_model(instance))
        instance.set_state(init_state)
        instance.save()



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations
from django.template.defaultfilters import slugify
from river.models.factories import ProceedingMetaObjectFactory
from river.services.proceeding_meta import ProceedingMetaService


def move_from_transition_to_meta(apps, schema_editor):
    # We can't import the Person model directly as it may be a newer
    # version than this migration expects. We use the historical version.
    ProceedingMeta = apps.get_model("river", "ProceedingMeta")
    for pm in ProceedingMeta.objects.all():
        pm.content_type = pm.transition.content_type
        pm.field = pm.transition.field
        pm.save()


class Migration(migrations.Migration):
    dependencies = [
        ('river', '0002_auto_20150916_0144'),
    ]

    operations = [
        migrations.RunPython(move_from_transition_to_meta),
    ]



from river.models.state import State
from river.services.object import ObjectService
from river.services.state import StateService
from river.tests.base_test import BaseTestCase

__author__ = 'ahmetdal'


class test__StateService(BaseTestCase):
    def test_get_available_states(self):
        self.initialize_normal_scenario()

        ObjectService.register_object(self.objects[0])
        ObjectService.register_object(self.objects[1])

        available_states = StateService.get_available_states(self.objects[0], self.user2, include_user=False)
        self.assertEqual(1, available_states.count())
        self.assertEqual(State.objects.get(label='s2'), available_states[0])
        available_states = StateService.get_available_states(self.objects[0], self.user2, include_user=True)
        self.assertEqual(0, available_states.count())
        available_states = StateService.get_available_states(self.objects[0], self.user2)
        self.assertEqual(0, available_states.count())

        available_states = StateService.get_available_states(self.objects[0], self.user3, include_user=False)
        self.assertEqual(1, available_states.count())
        self.assertEqual(State.objects.get(label='s2'), available_states[0])
        available_states = StateService.get_available_states(self.objects[0], self.user3, include_user=True)
        self.assertEqual(0, available_states.count())
        available_states = StateService.get_available_states(self.objects[0], self.user3)
        self.assertEqual(0, available_states.count())

        available_states = StateService.get_available_states(self.objects[0], self.user4, include_user=False)
        self.assertEqual(1, available_states.count())
        self.assertEqual(State.objects.get(label='s2'), available_states[0])
        available_states = StateService.get_available_states(self.objects[0], self.user4, include_user=True)
        self.assertEqual(0, available_states.count())
        available_states = StateService.get_available_states(self.objects[0], self.user4)
        self.assertEqual(0, available_states.count())

        available_states = StateService.get_available_states(self.objects[0], self.user1, include_user=False)
        self.assertEqual(1, available_states.count())
        self.assertEqual(State.objects.get(label='s2'), available_states[0])
        available_states = StateService.get_available_states(self.objects[0], self.user1, include_user=True)
        self.assertEqual(1, available_states.count())
        self.assertEqual(State.objects.get(label='s2'), available_states[0])

        available_states = StateService.get_available_states(self.objects[0], self.user1)
        self.assertEqual(1, available_states.count())
        self.assertEqual(State.objects.get(label='s2'), available_states[0])

    def test_get_initial_state(self):
        self.initialize_normal_scenario()

        self.assertEqual(State.objects.get(label='s1'), StateService.get_initial_state(self.content_type))

    def test_get_final_states(self):
        self.initialize_normal_scenario()

        self.assertListEqual(list(State.objects.filter(label__in=['s4.1', 's4.2', 's5.1', 's5.2'])), list(StateService.get_final_states(self.content_type)))



from __future__ import unicode_literals

from django.db import models
from django.db.models import CASCADE
from django.db.models.signals import m2m_changed, post_save
from django.utils.encoding import python_2_unicode_compatible
from django.utils.translation import ugettext_lazy as _

from river.config import app_config
from river.models.base_model import BaseModel
from river.models.managers.proceeding_meta import ProceedingMetaManager
from river.models.transition import Transition

__author__ = 'ahmetdal'


@python_2_unicode_compatible
class ProceedingMeta(BaseModel):
    class Meta:
        app_label = 'river'
        verbose_name = _("Proceeding Meta")
        verbose_name_plural = _("Proceeding Metas")
        unique_together = [('content_type', 'transition', 'order')]

    objects = ProceedingMetaManager()

    content_type = models.ForeignKey(app_config.CONTENT_TYPE_CLASS, verbose_name=_('Content Type'), on_delete=CASCADE)

    transition = models.ForeignKey(Transition, verbose_name=_('Transition'), on_delete=CASCADE)
    permissions = models.ManyToManyField(app_config.PERMISSION_CLASS, verbose_name=_('Permissions'), blank=True)
    groups = models.ManyToManyField(app_config.GROUP_CLASS, verbose_name=_('Groups'), blank=True)
    order = models.IntegerField(default=0, verbose_name=_('Order'), null=True)
    action_text = models.TextField(_("Action Text"), max_length=200, null=True, blank=True)

    parents = models.ManyToManyField('self', verbose_name='parents', related_name='children', symmetrical=False, db_index=True, blank=True)

    def natural_key(self):
        return self.content_type, self.transition, self.order

    def __str__(self):
        return 'Transition:%s, Permissions:%s, Groups:%s, Order:%s' % (
            self.transition, ','.join(self.permissions.values_list('name', flat=True)),
            ','.join(self.groups.values_list('name', flat=True)), self.order)


def post_group_change(sender, instance, *args, **kwargs):
    from river.services.proceeding import ProceedingService
    from river.models.proceeding import PENDING

    for proceeding_pending in instance.proceedings.filter(status=PENDING):
        ProceedingService.override_groups(proceeding_pending, instance.groups.all())


def post_permissions_change(sender, instance, *args, **kwargs):
    from river.models.proceeding import PENDING
    from river.services.proceeding import ProceedingService

    for proceeding_pending in instance.proceedings.filter(status=PENDING):
        ProceedingService.override_permissions(proceeding_pending, instance.permissions.all())


def post_save_model(sender, instance, *args, **kwargs):
    from river.services.proceeding_meta import ProceedingMetaService
    ProceedingMetaService.build_tree(instance)


m2m_changed.connect(post_group_change, sender=ProceedingMeta.groups.through)
m2m_changed.connect(post_permissions_change, sender=ProceedingMeta.permissions.through)

post_save.connect(post_save_model, sender=ProceedingMeta)



from django.db import models
from river.models.state import State

__author__ = 'ahmetdal'


class TransitionManager(models.Manager):
    def get_by_natural_key(self, source_state, destination_state):
        return self.get(source_state=State.objects.get_by_natural_key(source_state), destination_state=State.objects.get_by_natural_key(destination_state))



import logging
from river.handlers.backends.base import powerset

from river.handlers.backends.memory import MemoryHandlerBackend
from river.models.handler import Handler

__author__ = 'ahmetdal'

LOGGER = logging.getLogger(__name__)


class DatabaseHandlerBackend(MemoryHandlerBackend):

    def initialize_handlers(self):
        self.__register(Handler.objects.filter(enabled=True))

    def __register(self, handler_objs):
        handlers = []
        if handler_objs.exists():
            for handler in handler_objs:
                if handler.hash not in self.handlers:
                    module, method_name = handler.method.rsplit('.', 1)
                    try:
                        method = getattr(__import__(module, fromlist=[method_name]), method_name, None)
                        if method:
                            self.handlers[handler.hash] = method
                            handlers.append(method)
                            LOGGER.debug("Handler '%s' from database is registered initially from database as method '%s' and module '%s'. " % (handler.hash, method_name, module))
                        else:
                            LOGGER.warning("Handler '%s' from database can not be registered. Because method '%s' is not in module '%s'. " % (handler.hash, method_name, module))
                    except ImportError:
                        LOGGER.warning("Handler '%s' from database can not be registered. Because module '%s'  does not exists. " % (handler.hash, module))
        return handlers

    def register(self, handler_cls, handler, workflow_object, override=False, *args, **kwargs):
        hash = super(DatabaseHandlerBackend, self).register(handler_cls, handler, workflow_object, override=override, *args, **kwargs)
        handler_obj, created = Handler.objects.update_or_create(
            hash=hash,
            defaults={
                'method': '%s.%s' % (handler.__module__, handler.__name__),
                'handler_cls': '%s.%s' % (handler_cls.__module__, handler_cls.__name__),
            }
        )
        if created:
            LOGGER.debug("Handler '%s' is registered in database as method %s and module %s. " % (handler_obj.hash, handler.__name__, handler.__module__))
        else:
            LOGGER.warning("Handler '%s' is already registered in database as method %s and module %s. " % (handler_obj.hash, handler.__name__, handler.__module__))

        return hash

    def get_handlers(self, handler_cls, workflow_object, *args, **kwargs):
        handlers = super(DatabaseHandlerBackend, self).get_handlers(handler_cls, workflow_object, *args, **kwargs)
        if not handlers:
            hashes = []
            for c in powerset(kwargs.keys()):
                skwargs = {}
                for f in c:
                    skwargs[f] = kwargs.get(f)
                hash = self.get_handler_class(handler_cls).get_hash(workflow_object, **skwargs)
                hashes.append(self.get_handler_class_prefix(self.get_handler_class(handler_cls)) + hash)
            handlers = self.__register(Handler.objects.filter(hash__in=hashes))
        return handlers



from unittest import skip

from river.handlers.transition import PostTransitionHandler
from river.models.proceeding import Proceeding
from river.models.state import State
from river.services.object import ObjectService
from river.services.transition import TransitionService
from river.tests.base_test import BaseTestCase

__author__ = 'ahmetdal'


class test_TransitionHandler(BaseTestCase):
    @skip("workflow object is now required to register")
    def test_register_for_all(self):
        self.initialize_normal_scenario()
        self.test_args = None
        self.test_kwargs = None

        def test_handler(*args, **kwargs):
            self.test_args = args
            self.test_kwargs = kwargs

        ObjectService.register_object(self.objects[0])
        ObjectService.register_object(self.objects[1])

        PostTransitionHandler.register(test_handler)

        self.assertIsNone(self.test_args)
        self.assertIsNone(self.test_kwargs)

        TransitionService.proceed(self.objects[0], self.user1)

        self.assertEqual((), self.test_args)
        self.assertDictEqual(
            {
                'object': self.objects[0],
                'proceeding': Proceeding.objects.filter(meta__transition__source_state=State.objects.get(label='s1'),
                                                        meta__transition__destination_state=State.objects.get(label='s2'))[0],
                'source_state': State.objects.get(label='s1'),
                'destination_state': State.objects.get(label='s2')
            }, self.test_kwargs)

        # Proceeded but no transition
        TransitionService.proceed(self.objects[0], self.user2)

        self.assertEqual((), self.test_args)
        self.assertDictEqual(
            {
                'object': self.objects[0],
                'proceeding': Proceeding.objects.filter(meta__transition__source_state=State.objects.get(label='s1'),
                                                        meta__transition__destination_state=State.objects.get(label='s2'))[0],
                'source_state': State.objects.get(label='s1'),
                'destination_state': State.objects.get(label='s2')
            }, self.test_kwargs)

        TransitionService.proceed(self.objects[0], self.user3)

        self.assertEqual((), self.test_args)
        self.assertDictEqual(
            {
                'object': self.objects[0],
                'proceeding': Proceeding.objects.filter(meta__transition__source_state=State.objects.get(label='s2'),
                                                        meta__transition__destination_state=State.objects.get(label='s3'))[2],
                'source_state': State.objects.get(label='s2'),
                'destination_state': State.objects.get(label='s3')
            }, self.test_kwargs)

    def test_register_for_an_object(self):
        self.initialize_normal_scenario()
        self.test_args = None
        self.test_kwargs = None

        def test_handler(*args, **kwargs):
            self.test_args = args
            self.test_kwargs = kwargs

        ObjectService.register_object(self.objects[0])
        ObjectService.register_object(self.objects[1])

        PostTransitionHandler.register(test_handler, self.objects[1])

        self.assertIsNone(self.test_args)
        self.assertIsNone(self.test_kwargs)

        TransitionService.proceed(self.objects[0], self.user1)

        self.assertIsNone(self.test_args)
        self.assertIsNone(self.test_kwargs)

        TransitionService.proceed(self.objects[0], self.user2)

        self.assertIsNone(self.test_args)
        self.assertIsNone(self.test_kwargs)

        PostTransitionHandler.register(test_handler, self.objects[0])
        TransitionService.proceed(self.objects[0], self.user3)

        self.assertEqual((self.objects[0],), self.test_args)
        self.assertDictEqual(
            {
                'proceeding': Proceeding.objects.filter(meta__transition__source_state=State.objects.get(label='s2'),
                                                        meta__transition__destination_state=State.objects.get(label='s3'))[2],
                'source_state': State.objects.get(label='s2'),
                'destination_state': State.objects.get(label='s3')
            }, self.test_kwargs)

    def test_register_for_a_transition(self):
        self.initialize_normal_scenario()
        self.test_args = None
        self.test_kwargs = None

        def test_handler(*args, **kwargs):
            self.test_args = args
            self.test_kwargs = kwargs

        ObjectService.register_object(self.objects[0])
        ObjectService.register_object(self.objects[1])

        PostTransitionHandler.register(test_handler, self.objects[0], source_state=State.objects.get(label='s2'),
                                       destination_state=State.objects.get(label='s3'))

        self.assertIsNone(self.test_args)
        self.assertIsNone(self.test_kwargs)

        TransitionService.proceed(self.objects[0], self.user1)

        self.assertIsNone(self.test_args)
        self.assertIsNone(self.test_kwargs)

        TransitionService.proceed(self.objects[0], self.user2)

        self.assertIsNone(self.test_args)
        self.assertIsNone(self.test_kwargs)

        TransitionService.proceed(self.objects[0], self.user3)

        self.assertEqual((self.objects[0],), self.test_args)
        self.assertDictEqual(
            {
                'proceeding': Proceeding.objects.filter(meta__transition__source_state=State.objects.get(label='s2'),
                                                        meta__transition__destination_state=State.objects.get(label='s3'))[2],
                'source_state': State.objects.get(label='s2'),
                'destination_state': State.objects.get(label='s3')
            }, self.test_kwargs)



# -*- coding: utf-8 -*-
# Generated by Django 1.9.5 on 2016-05-24 04:39
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('river', '0005_auto_20160428_1818'),
    ]

    operations = [
        migrations.AlterField(
            model_name='proceedingmeta',
            name='groups',
            field=models.ManyToManyField(blank=True, null=True, to='auth.Group', verbose_name='Groups'),
        ),
        migrations.AlterField(
            model_name='proceedingmeta',
            name='order',
            field=models.IntegerField(default=0, null=True, verbose_name='Order'),
        ),
        migrations.AlterField(
            model_name='proceedingmeta',
            name='parents',
            field=models.ManyToManyField(blank=True, db_index=True, null=True, related_name='children', to='river.ProceedingMeta', verbose_name=b'parents'),
        ),
        migrations.AlterField(
            model_name='proceedingmeta',
            name='permissions',
            field=models.ManyToManyField(blank=True, to='auth.Permission', verbose_name='Permissions'),
        ),
        migrations.AlterField(
            model_name='transition',
            name='direction',
            field=models.SmallIntegerField(choices=[(0, 'Backward'), (1, 'Forward')], default=1, verbose_name='Transition Direction'),
        ),
    ]



from river.config import app_config
from river.handlers.backends.loader import load_handler_backend
from river.handlers.transition import PostTransitionHandler
from river.services.object import ObjectService
from river.tests.base_test import BaseTestCase

__author__ = 'ahmetdal'


def test_handler(*args, **kwargs):
    pass


class test_MemoryHandlerBackend(BaseTestCase):
    def setUp(self):
        super(test_MemoryHandlerBackend, self).setUp()
        self.initialize_normal_scenario()
        app_config.HANDLER_BACKEND_CLASS = 'river.handlers.backends.memory.MemoryHandlerBackend'
        self.handler_backend = load_handler_backend()
        self.handler_backend.handlers = {}

        ObjectService.register_object(self.objects[0])

    def test_register(self):
        self.assertFalse('%s.%s_object%s' % (
            PostTransitionHandler.__module__, PostTransitionHandler.__name__, self.objects[1].pk) in self.handler_backend.handlers)
        self.handler_backend.register(PostTransitionHandler, test_handler, self.objects[1])
        self.assertTrue(
            '%s.%s_object%s' % (PostTransitionHandler.__module__, PostTransitionHandler.__name__, self.objects[1].pk) in self.handler_backend.handlers)
        self.assertEqual(test_handler.__name__, list(self.handler_backend.handlers.values())[0].__name__)

    def test_get_handlers(self):
        self.handler_backend.register(PostTransitionHandler, test_handler, self.objects[1])
        handlers = self.handler_backend.get_handlers(PostTransitionHandler, self.objects[1])
        self.assertEqual(1, len(handlers))

        self.assertEqual(test_handler.__name__, handlers[0].__name__)



import factory
from django.contrib.contenttypes.models import ContentType
from django.test import TestCase

from river.tests.models import TestModel
from river.tests.models.factories import TestModelObjectFactory

__author__ = 'ahmetdal'


class BaseTestCase(TestCase):
    def setUp(self):
        super(BaseTestCase, self).setUp()
        print('%s is initialized' % self.__class__)

    def tearDown(self):
        super(BaseTestCase, self).tearDown()
        print('%s is finished' % self.__class__)

    def initialize_normal_scenario(self):
        from river.models.factories import \
            TransitionObjectFactory, \
            UserObjectFactory, \
            PermissionObjectFactory, \
            ProceedingMetaObjectFactory, \
            StateObjectFactory

        TransitionObjectFactory.reset_sequence(0)
        ProceedingMetaObjectFactory.reset_sequence(0)
        StateObjectFactory.reset_sequence(0)
        TestModel.objects.all().delete()

        self.content_type = ContentType.objects.get_for_model(TestModel)
        self.permissions = PermissionObjectFactory.create_batch(4)
        self.user1 = UserObjectFactory(user_permissions=[self.permissions[0]])
        self.user2 = UserObjectFactory(user_permissions=[self.permissions[1]])
        self.user3 = UserObjectFactory(user_permissions=[self.permissions[2]])
        self.user4 = UserObjectFactory(user_permissions=[self.permissions[3]])

        self.states = StateObjectFactory.create_batch(
            9,
            label=factory.Sequence(
                lambda n: "s%s" % str(n + 1) if n <= 4 else ("s4.%s" % str(n - 4) if n <= 6 else "s5.%s" % str(n - 6)))
        )
        self.transitions = TransitionObjectFactory.create_batch(8,
                                                                source_state=factory.Sequence(
                                                                    lambda n: self.states[n] if n <= 2 else (
                                                                        self.states[n - 1]) if n <= 4 else (
                                                                        self.states[n - 2] if n <= 6 else self.states[
                                                                            4])),
                                                                destination_state=factory.Sequence(
                                                                    lambda n: self.states[n + 1]))

        self.proceeding_metas = ProceedingMetaObjectFactory.create_batch(
            9,
            content_type=self.content_type,
            transition=factory.Sequence(lambda n: self.transitions[n] if n <= 1 else self.transitions[n - 1]),
            order=factory.Sequence(lambda n: 1 if n == 2 else 0)
        )

        for n, proceeding_meta in enumerate(self.proceeding_metas):
            proceeding_meta.permissions.add(self.permissions[n] if n <= 3 else self.permissions[3])

        self.objects = TestModelObjectFactory.create_batch(2)

    def initialize_circular_scenario(self):
        from river.models.factories import \
            TransitionObjectFactory, \
            UserObjectFactory, \
            PermissionObjectFactory, \
            ProceedingMetaObjectFactory, \
            StateObjectFactory

        TransitionObjectFactory.reset_sequence(0)
        ProceedingMetaObjectFactory.reset_sequence(0)
        StateObjectFactory.reset_sequence(0)
        TestModel.objects.all().delete()

        self.content_type = ContentType.objects.get_for_model(TestModel)
        self.permissions = PermissionObjectFactory.create_batch(4)
        self.user1 = UserObjectFactory(user_permissions=[self.permissions[0]])
        self.user2 = UserObjectFactory(user_permissions=[self.permissions[1]])
        self.user3 = UserObjectFactory(user_permissions=[self.permissions[2]])
        self.user4 = UserObjectFactory(user_permissions=[self.permissions[3]])


        self.open_state = StateObjectFactory(
            label='open'
        )
        self.in_progress_state = StateObjectFactory(
            label='in-progress'
        )

        self.resolved_state = StateObjectFactory(
            label='resolved'
        )
        self.re_opened_state = StateObjectFactory(
            label='re-opened'
        )

        self.closed_state = StateObjectFactory(
            label='closed'
        )

        self.transitions = [
            TransitionObjectFactory(source_state=self.open_state, destination_state=self.in_progress_state),
            TransitionObjectFactory(source_state=self.in_progress_state,
                                    destination_state=self.resolved_state),
            TransitionObjectFactory(source_state=self.resolved_state,
                                    destination_state=self.re_opened_state),
            TransitionObjectFactory(source_state=self.resolved_state, destination_state=self.closed_state),
            TransitionObjectFactory(source_state=self.re_opened_state, destination_state=self.in_progress_state)]

        self.proceeding_metas = ProceedingMetaObjectFactory.create_batch(
            5,
            content_type=self.content_type,
            transition=factory.Sequence(lambda n: self.transitions[n]),
            order=0
        )

        for n, proceeding_meta in enumerate(self.proceeding_metas):
            proceeding_meta.permissions.add(self.permissions[n] if n < len(self.permissions) else self.permissions[0])

        self.objects = TestModelObjectFactory.create_batch(2)



from unittest import skip

from river.handlers.completed import PostCompletedHandler
from river.models.transition import Transition
from river.services.object import ObjectService
from river.services.transition import TransitionService
from river.tests.base_test import BaseTestCase

__author__ = 'ahmetdal'


class test_CompletedHandler(BaseTestCase):
    def setUp(self):
        super(test_CompletedHandler, self).setUp()
        self.initialize_normal_scenario()
        transition = Transition.objects.get(source_state__label='s2', destination_state__label='s3')
        Transition.objects.filter(pk__gt=transition.pk).delete()

    @skip("workflow object is now required to register")
    def test_register_for_all(self):
        self.initialize_normal_scenario()
        self.test_args = None
        self.test_kwargs = None

        def test_handler(*args, **kwargs):
            self.test_args = args
            self.test_kwargs = kwargs

        ObjectService.register_object(self.objects[0])
        ObjectService.register_object(self.objects[1])

        PostCompletedHandler.register(test_handler)

        self.assertIsNone(self.test_args)
        self.assertIsNone(self.test_kwargs)

        TransitionService.proceed(self.objects[0], self.user1)

        self.assertIsNone(self.test_args)
        self.assertIsNone(self.test_kwargs)

        # Proceeded but no transition
        TransitionService.proceed(self.objects[0], self.user2)

        self.assertIsNone(self.test_args)
        self.assertIsNone(self.test_kwargs)

        TransitionService.proceed(self.objects[0], self.user3)

        self.assertEqual((), self.test_args)
        self.assertDictEqual(
            {
                'object': self.objects[0]
            }, self.test_kwargs)

    def test_register_for_an_object(self):
        self.test_args = None
        self.test_kwargs = None

        def test_handler(*args, **kwargs):
            self.test_args = args
            self.test_kwargs = kwargs

        ObjectService.register_object(self.objects[0])
        ObjectService.register_object(self.objects[1])

        PostCompletedHandler.register(test_handler, self.objects[0], 'my_field')

        self.assertIsNone(self.test_args)
        self.assertIsNone(self.test_kwargs)

        TransitionService.proceed(self.objects[0], self.user1)

        self.assertIsNone(self.test_args)
        self.assertIsNone(self.test_kwargs)

        # Proceeded but no transition
        TransitionService.proceed(self.objects[0], self.user2)

        self.assertIsNone(self.test_args)
        self.assertIsNone(self.test_kwargs)

        TransitionService.proceed(self.objects[0], self.user3)

        self.assertEqual((self.objects[0],), self.test_args)



import logging
from river.handlers.backends.base import BaseHandlerBackend, powerset

__author__ = 'ahmetdal'

LOGGER = logging.getLogger(__name__)


class MemoryHandlerBackend(BaseHandlerBackend):
    def __init__(self):
        self.handlers = {}

    def register(self, handler_cls, handler, workflow_object, override=False, *args, **kwargs):
        hash = self.get_handler_class_prefix(handler_cls) + handler_cls.get_hash(workflow_object, *args, **kwargs)
        if override or hash not in self.handlers:
            self.handlers[hash] = handler
            LOGGER.debug("Handler '%s'  is registered in memory as method '%s' and module '%s'. " % (hash, handler.__name__, handler.__module__))
        return hash

    def get_handlers(self, handler_cls, workflow_object, *args, **kwargs):
        handlers = []
        for c in powerset(kwargs.keys()):
            skwargs = {}
            for f in c:
                skwargs[f] = kwargs.get(f)
            hash = self.get_handler_class(handler_cls).get_hash(workflow_object, **skwargs)
            handler = self.handlers.get(self.get_handler_class_prefix(self.get_handler_class(handler_cls)) + hash)
            if handler:
                handlers.append(handler)
        return handlers



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations


class Migration(migrations.Migration):
    dependencies = [
        ('river', '0002_part_2'),
    ]

    operations = [

        migrations.RemoveField(
            model_name='transition',
            name='content_type',
        ),
        migrations.RemoveField(
            model_name='transition',
            name='field',
        ),
        migrations.AlterUniqueTogether(
            name='proceedingmeta',
            unique_together=set([('content_type', 'field', 'transition', 'order')]),
        ),
        migrations.AlterUniqueTogether(
            name='transition',
            unique_together=set([('source_state', 'destination_state')]),
        ),
        migrations.AddField(
            model_name='state',
            name='slug',
            field=models.SlugField(null=True, blank=True, unique=True),
            preserve_default=True
        ),
        migrations.AlterField(
            model_name='handler',
            name='date_created',
            field=models.DateTimeField(auto_now_add=True, verbose_name='Date Created', null=True),
        ),
        migrations.AlterField(
            model_name='handler',
            name='date_updated',
            field=models.DateTimeField(auto_now=True, verbose_name='Date Updated', null=True),
        ),
        migrations.AlterField(
            model_name='proceeding',
            name='date_created',
            field=models.DateTimeField(auto_now_add=True, verbose_name='Date Created', null=True),
        ),
        migrations.AlterField(
            model_name='proceeding',
            name='date_updated',
            field=models.DateTimeField(auto_now=True, verbose_name='Date Updated', null=True),
        ),
        migrations.AlterField(
            model_name='proceedingmeta',
            name='date_created',
            field=models.DateTimeField(auto_now_add=True, verbose_name='Date Created', null=True),
        ),
        migrations.AlterField(
            model_name='proceedingmeta',
            name='date_updated',
            field=models.DateTimeField(auto_now=True, verbose_name='Date Updated', null=True),
        ),
        migrations.AlterField(
            model_name='proceedingtrack',
            name='date_created',
            field=models.DateTimeField(auto_now_add=True, verbose_name='Date Created', null=True),
        ),
        migrations.AlterField(
            model_name='proceedingtrack',
            name='date_updated',
            field=models.DateTimeField(auto_now=True, verbose_name='Date Updated', null=True),
        ),
        migrations.AlterField(
            model_name='state',
            name='date_created',
            field=models.DateTimeField(auto_now_add=True, verbose_name='Date Created', null=True),
        ),
        migrations.AlterField(
            model_name='state',
            name='date_updated',
            field=models.DateTimeField(auto_now=True, verbose_name='Date Updated', null=True),
        ),
        migrations.AlterField(
            model_name='transition',
            name='date_created',
            field=models.DateTimeField(auto_now_add=True, verbose_name='Date Created', null=True),
        ),
        migrations.AlterField(
            model_name='transition',
            name='date_updated',
            field=models.DateTimeField(auto_now=True, verbose_name='Date Updated', null=True),
        ),
    ]



from django.contrib.contenttypes.models import ContentType

from django.db import models

from river.services.object import ObjectService

__author__ = 'ahmetdal'


class WorkflowObjectManager(models.Manager):
    def __init__(self, *args, **kwargs):
        super(WorkflowObjectManager, self).__init__()

    def get_objects_waiting_for_approval(self, user):
        return ObjectService.get_objects_waiting_for_approval(ContentType.objects.get_for_model(self.model), user)

    def get_object_count_waiting_for_approval(self, user):
        return ObjectService.get_object_count_waiting_for_approval(ContentType.objects.get_for_model(self.model), user)



from river.tests.models import TestModel

__author__ = 'ahmetdal'


class TestModelObjectFactory():
    def __init__(self):
        self.model = TestModel.objects.create(test_field="")

    @staticmethod
    def create_batch(size):
        for i in range(size):
            TestModel.objects.create(test_field=i)
        return TestModel.objects.all()



import logging

from django.dispatch import Signal

__author__ = 'ahmetdal'

pre_final = Signal(providing_args=["workflow_object", ])
post_final = Signal(providing_args=["workflow_object", ])

pre_transition = Signal(providing_args=["workflow_object", "source_state", "destination_state", "proceeding"])
post_transition = Signal(providing_args=["workflow_object", "source_state", "destination_state", "proceeding"])

pre_proceed = Signal(providing_args=["workflow_object", "proceeding"])
post_proceed = Signal(providing_args=["workflow_object", "proceeding"])

LOGGER = logging.getLogger(__name__)


class TransitionSignal(object):
    def __init__(self, status, workflow_object, proceeding):
        self.status = status
        self.workflow_object = workflow_object
        self.proceeding = proceeding

    def __enter__(self):
        if self.status:
            pre_transition.send(
                sender=TransitionSignal.__class__,
                workflow_object=self.workflow_object,
                source_state=self.proceeding.meta.transition.source_state,
                destination_state=self.proceeding.meta.transition.destination_state,
                proceeding=self.proceeding,
            )
            LOGGER.debug("Pre transition signal IS sent for workflow object %s for transition %s -> %s" % (
                self.workflow_object, self.proceeding.meta.transition.source_state.label, self.proceeding.meta.transition.destination_state.label))
        else:
            LOGGER.debug(
                "Pre transition signal IS NOT sent for workflow object %s. Although proceeding is occurred, not transition is occurred." % (
                    self.workflow_object))

    def __exit__(self, type, value, traceback):
        if self.status:
            post_transition.send(
                sender=TransitionSignal.__class__,
                workflow_object=self.workflow_object,
                source_state=self.proceeding.meta.transition.source_state,
                destination_state=self.proceeding.meta.transition.destination_state,
                proceeding=self.proceeding,
            )
            LOGGER.debug(
                "Post transition signal IS sent for workflow object %s for transition %s -> %s" % (
                    self.workflow_object, self.proceeding.meta.transition.source_state.label, self.proceeding.meta.transition.destination_state.label))
        else:
            LOGGER.debug(
                "Post transition signal IS NOT sent for workflow object %s. Although proceeding is occurred, not transition is occurred." % (
                    self.workflow_object))


class ProceedingSignal(object):
    def __init__(self, workflow_object, proceeding):
        self.workflow_object = workflow_object
        self.proceeding = proceeding

    def __enter__(self):
        pre_proceed.send(
            sender=ProceedingSignal.__class__,
            workflow_object=self.workflow_object,
            proceeding=self.proceeding,
        )
        LOGGER.debug("Pre proceeding signal IS sent for workflow object %s for transition %s -> %s" % (
            self.workflow_object, self.proceeding.meta.transition.source_state.label, self.proceeding.meta.transition.destination_state.label))

    def __exit__(self, type, value, traceback):
        post_proceed.send(
            sender=TransitionSignal.__class__,
            workflow_object=self.workflow_object,
            proceeding=self.proceeding,
        )
        LOGGER.debug("Post proceeding signal IS sent for workflow object %s for transition %s -> %s" % (
            self.workflow_object, self.proceeding.meta.transition.source_state.label, self.proceeding.meta.transition.destination_state.label))


class FinalSignal(object):
    def __init__(self, workflow_object):
        self.workflow_object = workflow_object
        self.status = self.workflow_object.on_final_state

    def __enter__(self):
        if self.status:
            pre_final.send(
                sender=FinalSignal.__class__,
                workflow_object=self.workflow_object,
            )
            LOGGER.debug("Pre final signal IS sent for workflow object %s for final state %s" % (self.workflow_object, self.workflow_object.get_state().label))

    def __exit__(self, type, value, traceback):
        if self.status:
            post_final.send(
                sender=TransitionSignal.__class__,
                workflow_object=self.workflow_object,
            )
            LOGGER.debug("Post final signal IS sent for workflow object %s for final state %s" % (self.workflow_object, self.workflow_object.get_state().label))



import logging
from datetime import datetime

from django.db.transaction import atomic

from river.models.proceeding import APPROVED
from river.services.proceeding import ProceedingService
from river.services.state import StateService
from river.signals import ProceedingSignal, TransitionSignal, FinalSignal
from river.utils.error_code import ErrorCode
from river.utils.exceptions import RiverException

__author__ = 'ahmetdal'


class TTSSignal(object):
    def __init__(self, status, pre_signal, post_signal, **kwargs):
        self.status = status
        self.pre_signal = pre_signal
        self.post_signal = post_signal
        self.kwargs = kwargs

    def __enter__(self):
        if self.status:
            self.pre_signal.send(
                sender=TTSSignal.__class__,
                **self.kwargs
            )

    def __exit__(self, type, value, traceback):
        if self.status:
            self.post_signal.send(
                sender=TTSSignal.__class__,
                **self.kwargs
            )


LOGGER = logging.getLogger(__name__)


class TransitionService(object):
    def __init__(self):
        pass

    @staticmethod
    @atomic
    def proceed(workflow_object, user, next_state=None, god_mod=False):

        def process(workflow_object, user, action, next_state=None, god_mod=False):
            proceedings = ProceedingService.get_available_proceedings(workflow_object, [workflow_object.get_state()], user=user, god_mod=god_mod)
            c = proceedings.count()
            if c == 0:
                raise RiverException(ErrorCode.NO_AVAILABLE_NEXT_STATE_FOR_USER, "There is no available state for destination for the user.")
            if c > 1:
                if next_state:
                    proceedings = proceedings.filter(meta__transition__destination_state=next_state)
                    if proceedings.count() == 0:
                        available_states = StateService.get_available_states(workflow_object, user)
                        raise RiverException(ErrorCode.INVALID_NEXT_STATE_FOR_USER, "Invalid state is given(%s). Valid states is(are) %s" % (
                            next_state.__str__(), ','.join([ast.__str__() for ast in available_states])))
                else:
                    raise RiverException(ErrorCode.NEXT_STATE_IS_REQUIRED,
                                         "State must be given when there are multiple states for destination")
            proceeding = proceedings[0]
            proceeding.status = action
            proceeding.transactioner = user
            proceeding.transaction_date = datetime.now()
            if workflow_object.proceeding:
                proceeding.previous = workflow_object.proceeding
            proceeding.save()

            return proceeding

        proceeding = process(workflow_object, user, APPROVED, next_state, god_mod)

        # Any other proceeding is left?
        required_proceedings = ProceedingService.get_available_proceedings(workflow_object, [workflow_object.get_state()], destination_state=next_state,
                                                                           god_mod=god_mod)

        transition_status = False
        if required_proceedings.count() == 0:
            workflow_object.set_state(proceeding.meta.transition.destination_state)
            transition_status = True

            # Next states should be PENDING back again if there is circle.
            ProceedingService.cycle_proceedings(workflow_object)
            # ProceedingService.get_next_proceedings(workflow_object).update(status=PENDING)

        with ProceedingSignal(workflow_object, proceeding), TransitionSignal(transition_status, workflow_object, proceeding), FinalSignal(workflow_object):
            workflow_object.save()

        LOGGER.debug("Workflow object %s is proceeded for next transition. Transition: %s -> %s" % (
            workflow_object, workflow_object.get_state().label, workflow_object.get_state()))



from django import forms
from django.contrib import admin
from django.contrib.contenttypes.models import ContentType

from river.models import ProceedingMeta
from river.models.fields.state import StateField
from river.services.object import ObjectService


def get_content_types():
    content_type_pks = []
    for ct in ContentType.objects.all():
        model = ct.model_class()
        if model is not None:
            for f in model._meta.fields:
                if type(f) is StateField:
                    content_type_pks.append(ct.pk)
    return content_type_pks


class ProceedingMetaForm(forms.ModelForm):
    content_type = forms.ModelChoiceField(queryset=ContentType.objects.none())

    class Meta:
        model = ProceedingMeta
        fields = ('content_type', 'transition', 'permissions', 'groups', 'order', 'action_text')

    def __init__(self, *args, **kwargs):
        self.declared_fields['content_type'].queryset = ContentType.objects.filter(pk__in=get_content_types())

        super(ProceedingMetaForm, self).__init__(*args, **kwargs)

    def save(self, *args, **kwargs):
        content_type = self.cleaned_data['content_type']
        field = ObjectService.get_field(content_type.model_class())
        instance = super(ProceedingMetaForm, self).save(commit=False)
        instance.field = field.name
        return super(ProceedingMetaForm, self).save(*args, **kwargs)


class ProceedingMetaAdmin(admin.ModelAdmin):
    form = ProceedingMetaForm


admin.site.register(ProceedingMeta, ProceedingMetaAdmin)



from datetime import datetime, timedelta

from django.contrib.auth.models import Group
from django.core.management import call_command

from river.models.factories import UserObjectFactory
from river.models.proceeding import Proceeding
from river.models.proceeding_meta import ProceedingMeta
from river.models.state import State
from river.services.object import ObjectService
from river.services.proceeding import ProceedingService
from river.tests.base_test import BaseTestCase
from river.tests.models import TestModelSlowCase1
from river.tests.models import TestModelSlowCase2
from river.tests.models.factories import TestModelObjectFactory

__author__ = 'ahmetdal'


class test_ProceedingService(BaseTestCase):
    def test_get_proceedings_object_waiting_for_approval_without_skip(self):
        self.initialize_normal_scenario()

        ObjectService.register_object(self.objects[0])
        ObjectService.register_object(self.objects[1])

        proceedings = ProceedingService.get_available_proceedings(self.objects[0], [self.objects[0].my_field], user=self.user1)
        self.assertEqual(1, proceedings.count())

        proceedings = ProceedingService.get_available_proceedings(self.objects[0], [self.objects[0].my_field], user=self.user2)
        self.assertEqual(0, proceedings.count())

        proceedings = ProceedingService.get_available_proceedings(self.objects[0], [self.objects[0].my_field], user=self.user3)
        self.assertEqual(0, proceedings.count())

        proceedings = ProceedingService.get_available_proceedings(self.objects[0], [self.objects[0].my_field], user=self.user4)
        self.assertEqual(0, proceedings.count())

    def test_get_next_proceedings(self):
        self.initialize_normal_scenario()

        ObjectService.register_object(self.objects[0])
        ObjectService.register_object(self.objects[1])

        proceedings = ProceedingService.get_next_proceedings(self.objects[0])
        self.assertEqual(9, proceedings.count())

        self.objects[0].proceed(self.user1)

        proceedings = ProceedingService.get_next_proceedings(self.objects[0])
        self.assertEqual(8, proceedings.count())

        self.objects[0].proceed(self.user2)

        # Two proceedings exist on same level
        proceedings = ProceedingService.get_next_proceedings(self.objects[0])
        self.assertEqual(7, proceedings.count())

        self.objects[0].proceed(self.user3)

        proceedings = ProceedingService.get_next_proceedings(self.objects[0])
        self.assertEqual(6, proceedings.count())

        self.objects[0].proceed(self.user4, next_state=State.objects.get(label='s4'))
        proceedings = ProceedingService.get_next_proceedings(self.objects[0])
        self.assertEqual(2, proceedings.count())

        self.objects[0].proceed(self.user4, next_state=State.objects.get(label='s4.1'))
        proceedings = ProceedingService.get_next_proceedings(self.objects[0])
        self.assertEqual(0, proceedings.count())

    def test_get_proceedings_object_waiting_for_approval_with_skip(self):
        self.initialize_normal_scenario()

        ObjectService.register_object(self.objects[0])
        ObjectService.register_object(self.objects[1])

        proceedings = self.objects[1].get_available_proceedings(self.user1)
        self.assertEqual(1, proceedings.count())
        self.assertEqual(State.objects.get(label='s2'), proceedings[0].meta.transition.destination_state)

        Proceeding.objects.filter(
            workflow_object=self.objects[1],
            meta__transition__destination_state=State.objects.get(label='s2')
        ).update(skip=True)

        proceedings = ProceedingService.get_available_proceedings(self.objects[1], [self.objects[1].my_field], user=self.user2)
        self.assertEqual(1, proceedings.count())
        self.assertEqual(State.objects.get(label='s3'), proceedings[0].meta.transition.destination_state)

        Proceeding.objects.filter(
            workflow_object=self.objects[1],
            meta__transition__destination_state=State.objects.get(label='s3')
        ).update(skip=True)

        proceedings = ProceedingService.get_available_proceedings(self.objects[1], [self.objects[1].my_field], user=self.user4)
        self.assertEqual(2, proceedings.count())
        self.assertEqual(State.objects.get(label='s4'), proceedings[0].meta.transition.destination_state)
        self.assertEqual(State.objects.get(label='s5'), proceedings[1].meta.transition.destination_state)

        Proceeding.objects.filter(
            workflow_object=self.objects[1],
            meta__transition__destination_state=State.objects.get(label='s4')
        ).update(skip=True)

        proceedings = ProceedingService.get_available_proceedings(self.objects[1], [self.objects[1].my_field], user=self.user4)
        self.assertEqual(3, proceedings.count())
        self.assertEqual(State.objects.get(label='s5'), proceedings[0].meta.transition.destination_state)
        self.assertEqual(State.objects.get(label='s4.1'), proceedings[1].meta.transition.destination_state)
        self.assertEqual(State.objects.get(label='s4.2'), proceedings[2].meta.transition.destination_state)

        Proceeding.objects.filter(
            workflow_object=self.objects[1],
            meta__transition__destination_state=State.objects.get(label='s4')
        ).update(skip=False)
        Proceeding.objects.filter(
            workflow_object=self.objects[1],
            meta__transition__destination_state=State.objects.get(label='s5')
        ).update(skip=True)

        proceedings = ProceedingService.get_available_proceedings(self.objects[1], [self.objects[1].my_field], user=self.user4)
        self.assertEqual(3, proceedings.count())
        self.assertEqual(State.objects.get(label='s4'), proceedings[0].meta.transition.destination_state)
        self.assertEqual(State.objects.get(label='s5.1'), proceedings[1].meta.transition.destination_state)
        self.assertEqual(State.objects.get(label='s5.2'), proceedings[2].meta.transition.destination_state)

        Proceeding.objects.filter(
            workflow_object=self.objects[1],
            meta__transition__destination_state__in=State.objects.filter(label__in=['s4', 's5'])
        ).update(skip=True)

        proceedings = ProceedingService.get_available_proceedings(self.objects[1],
                                                                  [self.objects[1].my_field], user=self.user4)
        self.assertEqual(4, proceedings.count())
        self.assertEqual(State.objects.get(label='s4.1'), proceedings[0].meta.transition.destination_state)
        self.assertEqual(State.objects.get(label='s4.2'), proceedings[1].meta.transition.destination_state)
        self.assertEqual(State.objects.get(label='s5.1'), proceedings[2].meta.transition.destination_state)
        self.assertEqual(State.objects.get(label='s5.2'), proceedings[3].meta.transition.destination_state)

        Proceeding.objects.filter(
            workflow_object=self.objects[1],
            meta__transition__destination_state__in=State.objects.filter(label__in=['s4.1', 's5.1'])
        ).update(skip=True)

        proceedings = ProceedingService.get_available_proceedings(self.objects[1], [self.objects[1].my_field], user=self.user4)
        self.assertEqual(2, proceedings.count())
        self.assertEqual(State.objects.get(label='s4.2'), proceedings[0].meta.transition.destination_state)
        self.assertEqual(State.objects.get(label='s5.2'), proceedings[1].meta.transition.destination_state)

    def test_cycle_proceedings(self):
        self.initialize_circular_scenario()

        ObjectService.register_object(self.objects[0])

        # No Cycle
        self.assertFalse(ProceedingService.cycle_proceedings(self.objects[0]))
        self.objects[0].proceed(user=self.user1, next_state=self.in_progress_state, god_mod=True)
        self.assertEqual(5, Proceeding.objects.filter(object_id=self.objects[0].pk).count())

        # No Cycle
        self.assertFalse(ProceedingService.cycle_proceedings(self.objects[0]))
        self.objects[0].proceed(user=self.user2, next_state=self.resolved_state, god_mod=True)
        self.assertEqual(5, Proceeding.objects.filter(object_id=self.objects[0].pk).count())

        # State is re-opened and cycle is detected. Transition in-progress to resolved proceeding is cloned
        self.assertFalse(ProceedingService.cycle_proceedings(self.objects[0]))
        self.objects[0].proceed(user=self.user3, next_state=self.re_opened_state, god_mod=True)
        self.assertEqual(6, Proceeding.objects.filter(object_id=self.objects[0].pk).count())
        self.assertEqual(ProceedingMeta.objects.get(transition__source_state=self.in_progress_state, transition__destination_state=self.resolved_state),
                         Proceeding.objects.filter(object_id=self.objects[0].pk).latest('date_created').meta)

        # There will be no cycling even if the method is invoked. Because cycling is done in proceeding.
        self.assertFalse(ProceedingService.cycle_proceedings(self.objects[0]))
        self.assertEqual(6, Proceeding.objects.filter(object_id=self.objects[0].pk).count())

        # State is in-progress and cycle is detected. Transition resolved to re-opened proceeding is cloned
        self.objects[0].proceed(user=self.user3, next_state=self.in_progress_state, god_mod=True)
        self.assertEqual(7, Proceeding.objects.filter(object_id=self.objects[0].pk).count())
        self.assertEqual(ProceedingMeta.objects.get(transition__source_state=self.resolved_state,
                                                    transition__destination_state=self.re_opened_state),
                         Proceeding.objects.filter(object_id=self.objects[0].pk).latest('date_created').meta)

        # State is resolved and cycle is detected. Transition re-opened to in-progress proceeding is cloned
        self.objects[0].proceed(user=self.user3, next_state=self.resolved_state, god_mod=True)
        self.assertEqual(8, Proceeding.objects.filter(object_id=self.objects[0].pk).count())
        self.assertEqual(ProceedingMeta.objects.get(transition__source_state=self.re_opened_state,
                                                    transition__destination_state=self.in_progress_state),
                         Proceeding.objects.filter(object_id=self.objects[0].pk).latest('date_created').meta)

        # State is re-opened and cycle is detected. Transition  in-progress to resolved proceeding is cloned
        self.assertFalse(ProceedingService.cycle_proceedings(self.objects[0]))
        self.objects[0].proceed(user=self.user3, next_state=self.re_opened_state, god_mod=True)
        self.assertEqual(9, Proceeding.objects.filter(object_id=self.objects[0].pk).count())
        self.assertEqual(ProceedingMeta.objects.get(transition__source_state=self.in_progress_state,
                                                    transition__destination_state=self.resolved_state),
                         Proceeding.objects.filter(object_id=self.objects[0].pk).latest('date_created').meta)

        # State is in-progress and cycle is detected. Transition resolved to re-opened proceeding is cloned
        self.objects[0].proceed(user=self.user3, next_state=self.in_progress_state, god_mod=True)
        self.assertEqual(10, Proceeding.objects.filter(object_id=self.objects[0].pk).count())
        self.assertEqual(ProceedingMeta.objects.get(transition__source_state=self.resolved_state,
                                                    transition__destination_state=self.re_opened_state),
                         Proceeding.objects.filter(object_id=self.objects[0].pk).latest('date_created').meta)

        # State is resolved and cycle is detected. Transition re-opened to in-progress proceeding is cloned
        self.objects[0].proceed(user=self.user3, next_state=self.resolved_state, god_mod=True)
        self.assertEqual(11, Proceeding.objects.filter(object_id=self.objects[0].pk).count())
        self.assertEqual(ProceedingMeta.objects.get(transition__source_state=self.re_opened_state,
                                                    transition__destination_state=self.in_progress_state),
                         Proceeding.objects.filter(object_id=self.objects[0].pk).latest('date_created').meta)

        # No Cycle for closed state.
        self.objects[0].proceed(user=self.user4, next_state=self.closed_state, god_mod=True)
        self.assertEqual(11, Proceeding.objects.filter(object_id=self.objects[0].pk).count())

    def test_get_proceedings_object_waiting_for_approval_slow_test(self):
        self.initialize_normal_scenario()
        self.objects = TestModelObjectFactory.create_batch(100)
        for o in self.objects:
            ObjectService.register_object(o)
        before = datetime.now()
        for o in self.objects:
            ProceedingService.get_available_proceedings(o, [o.my_field], user=self.user1)
        after = datetime.now()
        self.assertLess(after - before, timedelta(seconds=2))

    def test_get_proceedings_object_waiting_for_approval_slow_test_from_fixture(self):

        call_command('loaddata', 'river/fixtures/slow-case.json', verbosity=0)

        user1 = UserObjectFactory(groups=[Group.objects.get(pk=3)])

        objects1 = []
        for i in range(66703, 66723):
            objects1.append(TestModelSlowCase1.objects.create(pk=i))

        objects2 = []
        for i in range(8, 13):
            objects2.append(TestModelSlowCase2.objects.create(pk=i))

        for o in objects1:
            before = datetime.now()
            ProceedingService.get_available_proceedings(o, [o.status], user1)
            after = datetime.now()
            self.assertLess(after - before, timedelta(seconds=0.5))

        for o in objects2:
            before = datetime.now()
            ProceedingService.get_available_proceedings(o, [o.status], user1)
            after = datetime.now()
            self.assertLess(after - before, timedelta(seconds=0.5))



import logging

from river.handlers.backends.loader import handler_backend

__author__ = 'ahmetdal'

LOGGER = logging.getLogger(__name__)


class Handler(object):
    @classmethod
    def dispatch(cls, workflow_object, *args, **kwargs):
        LOGGER.debug("Handler %s is dispatched for workflow object %s" % (cls.__name__, workflow_object))
        kwargs.pop('signal', None)
        kwargs.pop('sender', None)

        for handler in handler_backend.get_handlers(cls, workflow_object, *args, **kwargs):
            handler(workflow_object, *args, **kwargs)
            LOGGER.debug("Handler %s for workflow object %s is found as method %s with args %s and kwargs %s" % (cls.__name__, workflow_object, handler.__name__, args, kwargs))

    @classmethod
    def register(cls, handler, workflow_object, override=False, *args, **kwargs):
        handler_backend.register(cls, handler, workflow_object, override, *args, **kwargs)

    @classmethod
    def get_hash(cls, workflow_object, *args, **kwargs):
        return 'object' + (str(workflow_object.pk) if workflow_object else '')



from __future__ import unicode_literals

from django.db import models
from django.db.models import CASCADE
from django.utils.encoding import python_2_unicode_compatible
from django.utils.translation import ugettext_lazy as _

from river.models.base_model import BaseModel
from river.models.managers.transition import TransitionManager
from river.models.state import State

__author__ = 'ahmetdal'

BACKWARD = 0
FORWARD = 1

DIRECTIONS = [
    (BACKWARD, _("Backward")),
    (FORWARD, _("Forward"))
]


@python_2_unicode_compatible
class Transition(BaseModel):
    class Meta:
        app_label = 'river'
        verbose_name = _("Transition")
        verbose_name_plural = _("Transitions")
        unique_together = (('source_state', 'destination_state'),)

    objects = TransitionManager()

    source_state = models.ForeignKey(State, verbose_name=_("Source State"), related_name='transitions_as_source', on_delete=CASCADE)
    destination_state = models.ForeignKey(State, verbose_name=_("Next State"), related_name='transitions_as_destination', on_delete=CASCADE)
    direction = models.SmallIntegerField(_("Transition Direction"), choices=DIRECTIONS, default=FORWARD)

    def natural_key(self):
        return self.source_state.slug, self.destination_state.slug

    def __str__(self):
        return '%s -> %s' % (self.source_state, self.destination_state)



from datetime import datetime
from django.conf import settings

from django.db import models
from django.utils.translation import ugettext_lazy as _

AUTH_USER_MODEL = getattr(settings, 'AUTH_USER_MODEL', 'auth.User')


class BaseModel(models.Model):
    date_created = models.DateTimeField(_('Date Created'), null=True, blank=True, auto_now_add=True)
    date_updated = models.DateTimeField(_('Date Updated'), null=True, blank=True, auto_now=True)

    class Meta:
        app_label = 'river'
        abstract = True

    def details(self):
        return {'pk': self.pk}

    def authenticate_header(self, request):
        return None



# -*- coding: utf-8 -*-
# Generated by Django 1.10.5 on 2017-01-24 11:25
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('river', '0008_auto_20161017_0916'),
    ]

    operations = [
        migrations.AlterField(
            model_name='proceedingmeta',
            name='groups',
            field=models.ManyToManyField(blank=True, to='auth.Group', verbose_name='Groups'),
        ),
        migrations.AlterField(
            model_name='proceedingmeta',
            name='parents',
            field=models.ManyToManyField(blank=True, db_index=True, related_name='children', to='river.ProceedingMeta', verbose_name='parents'),
        ),
    ]



from river.handlers.handler import Handler
from river.signals import pre_final, post_final

__author__ = 'ahmetdal'


class PreCompletedHandler(Handler):
    pass


class PostCompletedHandler(Handler):
    pass


pre_final.connect(PreCompletedHandler.dispatch)
post_final.connect(PostCompletedHandler.dispatch)



from river.models.proceeding import Proceeding
from river.services.proceeding import ProceedingService
from river.utils.error_code import ErrorCode
from river.utils.exceptions import RiverException

__author__ = 'ahmetdal'


# noinspection PyClassHasNoInit
class ObjectService:
    @staticmethod
    def register_object(workflow_object):
        proceedings = Proceeding.objects.filter(workflow_object=workflow_object)
        if proceedings.count() == 0:
            ProceedingService.init_proceedings(workflow_object)

    @staticmethod
    def get_objects_waiting_for_approval(content_type, user):
        object_pks = []
        WorkflowObjectClass = content_type.model_class()
        for workflow_object in WorkflowObjectClass.objects.all():
            proceedings = ProceedingService.get_available_proceedings(workflow_object, [workflow_object.get_state()], user=user)
            if proceedings.count():
                object_pks.append(workflow_object.pk)
        return WorkflowObjectClass.objects.filter(pk__in=object_pks)

    @staticmethod
    def get_object_count_waiting_for_approval(content_type, user):
        return ObjectService.get_objects_waiting_for_approval(content_type, user).count()

    @staticmethod
    def is_workflow_completed(workflow_object):
        return Proceeding.objects.filter(workflow_object=workflow_object, meta__transition__source_state=workflow_object.get_state()).count() == 0

    @staticmethod
    def get_field(workflow_class):
        from river.models.fields.state import StateField

        field = next((f for f in workflow_class._meta.fields if type(f) is StateField), None)

        if not field:
            raise RiverException(ErrorCode.NO_STATE_FIELD, "There is no state field in model class of given instance")

        return field



from django.conf import settings
from django.contrib.auth.models import Permission, Group
from django.contrib.contenttypes.models import ContentType

from django.conf import settings

__author__ = 'ahmetdal'


class RiverConfig(object):
    # from settings
    prefix = 'RIVER'

    def get_with_prefix(self, config):
        return '%s_%s' % (self.prefix, config)

    def load(self):
        self.CONTENT_TYPE_CLASS = getattr(settings, self.get_with_prefix('CONTENT_TYPE_CLASS'), ContentType)
        self.USER_CLASS = getattr(settings, self.get_with_prefix('USER_CLASS'), settings.AUTH_USER_MODEL)
        self.PERMISSION_CLASS = getattr(settings, self.get_with_prefix('PERMISSION_CLASS'), Permission)
        self.GROUP_CLASS = getattr(settings, self.get_with_prefix('GROUP_CLASS'), Group)
        self.HANDLER_BACKEND = getattr(settings, self.get_with_prefix('HANDLER_BACKEND'), {'backend': 'river.handlers.backends.memory.MemoryHandlerBackend'})

        # Generated
        self.HANDLER_BACKEND_CLASS = self.HANDLER_BACKEND.get('backend')
        self.HANDLER_BACKEND_CONFIG = self.HANDLER_BACKEND.get('config', {})


app_config = RiverConfig()
app_config.load()



from django.db import models

__author__ = 'ahmetdal'


class ProceedingMetaManager(models.Manager):
    def get_by_natural_key(self, content_type, transition, order):
        return self.get(content_type=content_type, transition=transition, order=order)



from copy import deepcopy

import mock
from django.db import models

from river.models.fields.state import StateField, _get_cls_identifier, class_field_rl
from river.tests.base_test import BaseTestCase
from river.tests.models.testmodel import TestModel
from river.utils.error_code import ErrorCode
from river.utils.exceptions import RiverException

__author__ = 'ahmetdal'


class test_StateFIeld(BaseTestCase):
    def test_injections(self):
        self.assertTrue(hasattr(TestModel, 'proceedings'))
        self.assertTrue(hasattr(TestModel, 'proceeding'))
        self.assertTrue(hasattr(TestModel, 'is_workflow_completed'))
        self.assertTrue(hasattr(TestModel, 'proceed'))
        self.assertTrue(hasattr(TestModel, 'on_initial_state'))
        self.assertTrue(hasattr(TestModel, 'on_final_state'))
        self.assertTrue(hasattr(TestModel, 'get_initial_state'))
        self.assertTrue(hasattr(TestModel, 'get_available_proceedings'))
        self.assertTrue(hasattr(TestModel, 'initial_proceedings'))
        self.assertTrue(hasattr(TestModel, 'final_proceedings'))
        self.assertTrue(hasattr(TestModel, 'next_proceedings'))
        self.assertTrue(hasattr(TestModel, 'get_state'))
        self.assertTrue(hasattr(TestModel, 'set_state'))
        self.assertTrue(hasattr(TestModel, 'objects'))

    @mock.patch('django.db.models.fields.related.ForeignObject.contribute_to_class', mock.MagicMock())
    @mock.patch('river.models.fields.state.StateField._StateField__add_to_class', mock.MagicMock())
    def test_triggering_multiple_time(self):
        class TestModelForTriggeringMultipleTime(models.Model):
            pass

        field = StateField()
        class_field_rl_cp = deepcopy(class_field_rl)

        field.contribute_to_class(TestModelForTriggeringMultipleTime, 'status')

        class_field_rl_cp.update({_get_cls_identifier(TestModelForTriggeringMultipleTime): 'status'})
        self.assertDictEqual(class_field_rl_cp, class_field_rl)

        field.contribute_to_class(TestModelForTriggeringMultipleTime, 'status')

        self.assertDictEqual(class_field_rl_cp, class_field_rl)

        try:
            field.contribute_to_class(TestModelForTriggeringMultipleTime, 'status2')
            self.assertFalse(True, "River exception with error code %d, must have been raised")
        except RiverException as re:
            self.assertEqual(ErrorCode.MULTIPLE_STATE_FIELDS, re.code)
        else:
            self.assertFalse(True, "River exception with error code %d, must have been raised")



from river.models.state import State
from river.models.proceeding import Proceeding, APPROVED

from river.services.object import ObjectService
from river.services.transition import TransitionService
from river.utils.error_code import ErrorCode
from river.utils.exceptions import RiverException
from river.tests.base_test import BaseTestCase

__author__ = 'ahmetdal'


class test__TransitionService(BaseTestCase):
    def test_proceed(self):
        self.initialize_normal_scenario()

        ObjectService.register_object(self.objects[0])
        ObjectService.register_object(self.objects[1])
        # ####################
        # STATE 1 - STATE 2
        # Only User1(2001) can proceed and after his proceed state must be changed to STATE 2
        # ###################

        # Proceeded by user has no required permission for this transition

        try:
            TransitionService.proceed(self.objects[0], self.user2)
            self.fail('Exception was expected')
        except RiverException as e:
            self.assertEqual(str(e), 'There is no available state for destination for the user.')
            self.assertEqual(ErrorCode.NO_AVAILABLE_NEXT_STATE_FOR_USER, e.code)

        try:
            TransitionService.proceed(self.objects[0], self.user2)
            self.fail('Exception was expected')
        except RiverException as e:
            self.assertEqual(str(e), 'There is no available state for destination for the user.')
            self.assertEqual(ErrorCode.NO_AVAILABLE_NEXT_STATE_FOR_USER, e.code)


        # Proceeded by user has no required permission for this transition

        try:
            TransitionService.proceed(self.objects[0], self.user3)
            self.fail('Exception was expected')
        except RiverException as e:
            self.assertEqual(str(e), 'There is no available state for destination for the user.')
            self.assertEqual(ErrorCode.NO_AVAILABLE_NEXT_STATE_FOR_USER, e.code)

        try:
            TransitionService.proceed(self.objects[0], self.user3)
            self.fail('Exception was expected')
        except RiverException as e:
            self.assertEqual(str(e), 'There is no available state for destination for the user.')
            self.assertEqual(ErrorCode.NO_AVAILABLE_NEXT_STATE_FOR_USER, e.code)


        # Proceeded by user has no required permission for this transition

        try:
            TransitionService.proceed(self.objects[0], self.user4)
            self.fail('Exception was expected')
        except RiverException as e:
            self.assertEqual(str(e), 'There is no available state for destination for the user.')
            self.assertEqual(ErrorCode.NO_AVAILABLE_NEXT_STATE_FOR_USER, e.code)

        try:
            TransitionService.proceed(self.objects[0], self.user4)
            self.fail('Exception was expected')
        except RiverException as e:
            self.assertEqual(str(e), 'There is no available state for destination for the user.')
            self.assertEqual(ErrorCode.NO_AVAILABLE_NEXT_STATE_FOR_USER, e.code)

        # Proceeded by user has required permission for this transition

        self.assertEqual(State.objects.get(label='s1'), self.objects[0].get_state())

        TransitionService.proceed(self.objects[0], self.user1)

        self.assertEqual(State.objects.get(label='s2'), self.objects[0].get_state())

        proceedings = Proceeding.objects.filter(
            workflow_object=self.objects[0],
            status=APPROVED,
            meta__transition__source_state__label='s1',
            meta__transition__destination_state__label='s2'
        )
        self.assertEqual(1, proceedings.count())
        self.assertIsNotNone(proceedings[0].transactioner)
        self.assertEqual(self.user1, proceedings[0].transactioner)
        self.assertIsNotNone(proceedings[0].transaction_date)

        try:
            TransitionService.proceed(self.objects[0], self.user1)
            self.fail('Exception was expected')
        except RiverException as e:
            self.assertEqual(str(e), 'There is no available state for destination for the user.')
            self.assertEqual(ErrorCode.NO_AVAILABLE_NEXT_STATE_FOR_USER, e.code)


        # ####################
        # STATE 2 - STATE 3
        # User2(2002) is first proceeder and User3(2003) is second proceeder. This must be done with turn. After proceeding is done, state is gonna be changed to STATE 3
        # ####################

        # Proceeded by user has no required permission for this transition
        try:
            TransitionService.proceed(self.objects[0], self.user1)
            self.fail('Exception was expected')
        except RiverException as e:
            self.assertEqual(str(e), 'There is no available state for destination for the user.')
            self.assertEqual(ErrorCode.NO_AVAILABLE_NEXT_STATE_FOR_USER, e.code)

        # Proceeded by user has no required permission for this transition
        try:
            TransitionService.proceed(self.objects[0], self.user4)
            self.fail('Exception was expected')
        except RiverException as e:
            self.assertEqual(str(e), 'There is no available state for destination for the user.')
            self.assertEqual(ErrorCode.NO_AVAILABLE_NEXT_STATE_FOR_USER, e.code)

        # Turn is User2(2002)s, not User3(2003)s. After User2(2002) proceeded, User3(2003) can proceed.
        try:
            TransitionService.proceed(self.objects[0], self.user3)
            self.fail('Exception was expected')
        except RiverException as e:
            self.assertEqual(str(e), 'There is no available state for destination for the user.')
            self.assertEqual(ErrorCode.NO_AVAILABLE_NEXT_STATE_FOR_USER, e.code)


        # Proceeded by two user has required permission for this transition to get next state (order is user2(2002),user3(2003)).

        self.assertEqual(State.objects.get(label='s2'), self.objects[0].get_state())
        TransitionService.proceed(self.objects[0], self.user2)

        self.assertEqual(State.objects.get(label='s2'), self.objects[0].get_state())

        proceedings = Proceeding.objects.filter(
            workflow_object=self.objects[0],
            status=APPROVED,
            meta__transition__source_state__label='s2',
            meta__transition__destination_state__label='s3'
        )
        self.assertEqual(1, proceedings.count())
        self.assertIsNotNone(proceedings[0].transactioner)
        self.assertEqual(self.user2, proceedings[0].transactioner)
        self.assertIsNotNone(proceedings[0].transaction_date)

        try:
            TransitionService.proceed(self.objects[0], self.user2)
            self.fail('Exception was expected')
        except RiverException as e:
            self.assertEqual(str(e), 'There is no available state for destination for the user.')
            self.assertEqual(ErrorCode.NO_AVAILABLE_NEXT_STATE_FOR_USER, e.code)

        self.assertEqual(State.objects.get(label='s2'), self.objects[0].get_state())

        TransitionService.proceed(self.objects[0], self.user3)

        self.assertEqual(State.objects.get(label='s3'), self.objects[0].get_state())

        proceedings = Proceeding.objects.filter(
            workflow_object=self.objects[0],
            status=APPROVED,
            meta__transition__source_state__label='s2',
            meta__transition__destination_state__label='s3'
        )
        self.assertEqual(2, proceedings.count())
        self.assertIsNotNone(proceedings[0].transactioner)
        self.assertIsNotNone(proceedings[1].transactioner)
        self.assertEqual(self.user2, proceedings[0].transactioner)
        self.assertEqual(self.user3, proceedings[1].transactioner)
        self.assertIsNotNone(proceedings[0].transaction_date)
        self.assertIsNotNone(proceedings[1].transaction_date)

        try:
            TransitionService.proceed(self.objects[0], self.user3)
            self.fail('Exception was expected')
        except RiverException as e:
            self.assertEqual(str(e), 'There is no available state for destination for the user.')
            self.assertEqual(ErrorCode.NO_AVAILABLE_NEXT_STATE_FOR_USER, e.code)



        # ####################
        # STATE 3 - STATE 4 or STATE 5
        # Only User4(2004) can proceed by giving the exact next state and after his proceed with his state must be changed to STATE 4 or STATE 5
        # ###################

        # Proceeded by user has no required permission for this transition
        try:
            TransitionService.proceed(self.objects[0], self.user1)
            self.fail('Exception was expected')
        except RiverException as e:
            self.assertEqual(str(e), 'There is no available state for destination for the user.')
            self.assertEqual(ErrorCode.NO_AVAILABLE_NEXT_STATE_FOR_USER, e.code)

        # Proceeded by user has no required permission for this transition
        try:
            TransitionService.proceed(self.objects[0], self.user2)
            self.fail('Exception was expected')
        except RiverException as e:
            self.assertEqual(str(e), 'There is no available state for destination for the user.')
            self.assertEqual(ErrorCode.NO_AVAILABLE_NEXT_STATE_FOR_USER, e.code)

        # Proceeded by user has no required permission for this transition
        try:
            TransitionService.proceed(self.objects[0], self.user3)
            self.fail('Exception was expected')
        except RiverException as e:
            self.assertEqual(str(e), 'There is no available state for destination for the user.')
            self.assertEqual(ErrorCode.NO_AVAILABLE_NEXT_STATE_FOR_USER, e.code)


        # There are STATE 4 and STATE 5 as next. State must be given to switch
        try:
            TransitionService.proceed(self.objects[0], self.user4)
            self.fail('Exception was expected')
        except RiverException as e:
            self.assertEqual(str(e), 'State must be given when there are multiple states for destination')
            self.assertEqual(ErrorCode.NEXT_STATE_IS_REQUIRED, e.code)


        # There are STATE 4 and STATE 5 as next. State among STATE 4 and STATE 5 must be given to switch, not other state
        try:
            TransitionService.proceed(self.objects[0], self.user4, next_state=State.objects.get(label='s3'))
            self.fail('Exception was expected')
        except RiverException as e:
            self.assertEqual(str(e),
                             "Invalid state is given(%s). Valid states is(are) %s" % (
                                 State.objects.get(label='s3').__str__(), ','.join([ast.__str__() for ast in State.objects.filter(label__in=['s4', 's5'])])))
            self.assertEqual(ErrorCode.INVALID_NEXT_STATE_FOR_USER, e.code)




        # There are STATE 4 and STATE 5 as next. After one of them is given to proceeding, the state must be switch to it immediately.
        self.assertEqual(State.objects.get(label='s3'), self.objects[0].get_state())

        TransitionService.proceed(self.objects[0], self.user4, next_state=State.objects.get(label='s5'))

        self.assertEqual(State.objects.get(label='s5'), self.objects[0].get_state())

        proceedings = Proceeding.objects.filter(
            workflow_object=self.objects[0],
            status=APPROVED,
            meta__transition__source_state__label='s3',
            meta__transition__destination_state__label='s5'
        )
        self.assertEqual(1, proceedings.count())
        self.assertIsNotNone(proceedings[0].transactioner)
        self.assertEqual(self.user4, proceedings[0].transactioner)
        self.assertIsNotNone(proceedings[0].transaction_date)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations
from river.models.factories import ProceedingMetaObjectFactory
from river.services.proceeding_meta import ProceedingMetaService


class Migration(migrations.Migration):
    dependencies = [
        ('river', '0002_part_3'),
    ]

    operations = [
        migrations.AddField(
            model_name='proceedingmeta',
            name='parents',
            field=models.ManyToManyField(related_name='children', verbose_name=b'parents', to='river.ProceedingMeta', db_index=True),
        )
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.9.5 on 2016-04-28 18:18
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion
import mptt.fields


class Migration(migrations.Migration):

    dependencies = [
        ('river', '0004_data_fix'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='proceedingtrack',
            name='previous_track',
        ),
        migrations.RemoveField(
            model_name='proceedingtrack',
            name='proceeding',
        ),
        migrations.AddField(
            model_name='proceeding',
            name='cloned',
            field=models.BooleanField(default=False, verbose_name='Cloned?'),
        ),
        migrations.AddField(
            model_name='proceeding',
            name='previous',
            field=mptt.fields.TreeOneToOneField(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='next_proceeding', to='river.Proceeding', verbose_name='Previous Proceeding'),
        ),
        migrations.DeleteModel(
            name='ProceedingTrack',
        ),
    ]



# -*- coding: utf-8 -*-
from __future__ import unicode_literals
import uuid

from django.db import models, migrations
from django.db.models import CASCADE
from django.utils.text import slugify


class Migration(migrations.Migration):
    dependencies = [
        ('river', '0001_initial'),
    ]

    operations = [
        migrations.AddField(
            model_name='proceedingmeta',
            name='content_type',
            field=models.ForeignKey(default=1, verbose_name='Content Type', to='contenttypes.ContentType', on_delete=CASCADE),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name='proceedingmeta',
            name='field',
            field=models.CharField(default='state', max_length=200, verbose_name='Field'),
            preserve_default=False,
        ),
    ]



from itertools import combinations, chain

__author__ = 'ahmetdal'


def powerset(iterable):
    xs = list(iterable)
    # note we return an iterator rather than a list
    return chain.from_iterable(combinations(xs, n) for n in range(len(xs) + 1))


class BaseHandlerBackend(object):
    def register(self, handler_cls, handler, workflow_object, override=False, *args, **kwargs):
        raise NotImplementedError()

    def get_handlers(self, handler_cls, handler, workflow_object, *args, **kwargs):
        raise NotImplementedError()

    def get_handler_class(self, handler_cls):
        if isinstance(handler_cls, str):
            module, cls = handler_cls.rsplit('.', 1)
            handler_cls = getattr(__import__(module, fromlist=[cls]), cls)
        return handler_cls

    def get_handler_class_prefix(self, handler_cls):
        return '%s.%s_' % (handler_cls.__module__, handler_cls.__name__)



import logging

from django.apps import AppConfig
from django.db.utils import OperationalError, ProgrammingError


__author__ = 'ahmetdal'

LOGGER = logging.getLogger(__name__)


class RiverApp(AppConfig):
    name = 'river'
    label = 'river'

    def ready(self):

        from river.handlers.backends.database import DatabaseHandlerBackend
        from river.handlers.backends.loader import handler_backend        

        if isinstance(handler_backend, DatabaseHandlerBackend):
            try:
                self.get_model('Handler').objects.exists()
                handler_backend.initialize_handlers()
            except (OperationalError, ProgrammingError):
                LOGGER.debug('Database handlers are not registered. Because database is not created yet.')
        LOGGER.debug('RiverApp is loaded.')



from django.db import models

from river.models.fields.state import StateField

__author__ = 'ahmetdal'


class TestModel(models.Model):
    test_field = models.CharField(max_length=50, null=True, blank=True)
    my_field = StateField()


class TestModelWithoutStateField(models.Model):
    test_field = models.CharField(max_length=50, null=True, blank=True)


class TestModelSlowCase1(models.Model):
    status = StateField()


class TestModelSlowCase2(models.Model):
    status = StateField()



# -*- coding: utf-8 -*-
# Generated by Django 1.10 on 2016-10-17 09:16
from __future__ import unicode_literals

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('river', '0007_auto_20161015_0959'),
    ]

    operations = [
        migrations.AlterUniqueTogether(
            name='proceedingmeta',
            unique_together=set([('content_type', 'transition', 'order')]),
        ),
        migrations.RemoveField(
            model_name='proceeding',
            name='field',
        ),
        migrations.RemoveField(
            model_name='proceedingmeta',
            name='field',
        ),
    ]



from river.models.proceeding import Proceeding
from river.services.object import ObjectService
from river.services.state import StateService
from river.tests.base_test import BaseTestCase
from river.tests.models.testmodel import TestModel, TestModelWithoutStateField
from river.utils.error_code import ErrorCode
from river.utils.exceptions import RiverException

__author__ = 'ahmetdal'


# noinspection PyPep8Naming
class test_ObjectService(BaseTestCase):
    def test_init(self):
        self.initialize_normal_scenario()

        ObjectService.register_object(self.objects[0])
        ObjectService.register_object(self.objects[1])

        initial_state = StateService.get_initial_state(self.content_type)

        self.assertEqual(initial_state, self.objects[0].get_state())
        self.assertEqual(18, Proceeding.objects.count())

    def test_get_objects_waiting_for_approval_for_user(self):
        self.initialize_normal_scenario()

        on_approval_objects = ObjectService.get_objects_waiting_for_approval(self.content_type, self.user1)
        self.assertEqual(2, on_approval_objects.count())
        self.assertEqual(self.objects[0], on_approval_objects[0])

        on_approval_objects = ObjectService.get_objects_waiting_for_approval(self.content_type, self.user2)
        self.assertEqual(0, on_approval_objects.count())

        on_approval_objects = ObjectService.get_objects_waiting_for_approval(self.content_type, self.user3)
        self.assertEqual(0, on_approval_objects.count())

        on_approval_objects = ObjectService.get_objects_waiting_for_approval(self.content_type, self.user4)
        self.assertEqual(0, on_approval_objects.count())

    def test_get_field(self):
        field = ObjectService.get_field(TestModel)
        self.assertEqual('my_field', field.name)

        try:
            ObjectService.get_field(TestModelWithoutStateField)
            self.assertFalse(True, "Trying getting field from a model does not contains state field should have thrown the error with code '8'")
        except RiverException as re:
            self.assertEqual(ErrorCode.NO_STATE_FIELD, re.code)



from django.test import TestCase
from river.models.state import State

__author__ = 'ahmetdal'


class test__StateModel(TestCase):
    def test_on_pre_save_without_slug(self):
        expected_slug = 'test-label'
        state = State(label='Test Label')
        state.save()
        self.assertEqual(expected_slug, state.slug)

    def test_on_pre_save_with_wrong_formatted_slug(self):
        expected_slug = 'test-slug'
        state = State(label='Test Label', slug='Test Slug')
        state.save()
        self.assertEqual(expected_slug, state.slug)
